# Copyright CNRS/Inria/UCA
# Contributor(s): Eric Debreuve (since 2019)
#
# eric.debreuve@cnrs.fr
#
# This software is governed by the CeCILL  license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".
#
# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.
#
# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.
#
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

import re as rgex
from typing import Any, Dict, Final, List, Sequence, Tuple, Union

import numpy as nmpy
from im_tools_36.type.image import intensity_h, pix_value_h, spectrum_h
from scipy import ndimage as spim


array_t = nmpy.ndarray


GRAYSCALE_MODE_MARKER: Final = "g"
VALUE_GUESSING_REGEX: Final = rgex.compile(
    r"([" + GRAYSCALE_MODE_MARKER + r"m]uess)([0-9]*)"
)


def AutoCropped(
    img: array_t,
    /,
    *,
    force_color: bool = False,
    axis: Union[int, Sequence[int]] = None,
    value: Union[pix_value_h, str] = 0,
    tolerance: float = 0.0,
    return_value: bool = False,
    return_offsets: bool = False,
) -> Union[
    array_t,
    Tuple[array_t, array_t],
    Tuple[array_t, Tuple[int, ...]],
    Tuple[array_t, array_t, Tuple[int, ...]],
]:
    """
    The image is considered to be grayscale (or scalar) if "force_color" is False and "value" is a number or a string of
    the format "guess[some_int]" (see below). If "force_color" is True, or "value" is a sequence of numbers or a string
    of the format "muess[some_int]" (see below), the image is treated as a multichannel image (or multi/hyperspectral,
    or vector-valued, or color (whatever the color space is)), the channel dimension being the last one. The length of
    "value" must therefore be equal to this dimension. Note that for a multichannel image, "value" can actually be a
    number, in which case it will be treated as a tuple of this number repeated as many times as the image length in its
    last dimension.
    If "value" is a string, it can be:
    - "guess[some_int]":
    - "muess[some_int]": ("guess" with "g"->"m" as if it meant "Grayscale"->"Multichannel")
    some_int: The border width used to look for the most frequent value (grayscale) or value sequence (multichannel). It
    is 1 by default.

    tolerance: Fraction of the full dynamic of the image, if grayscale, or the largest full dynamic among the channels
    if multichannel. The latter could be replaced with the full vectorial dynamic using the L2 norm for example.
    However, this would require computing the diameter of the vector point cloud. This might be implemented in the
    future.
    In theory, the tolerance should belong to [0,1] for both grayscale and multichannel. However, due to the full
    dynamic computation method for multichannel, the useful range goes beyond 1 in this case.

    return_value: If true, the function returns the value that has been cropped; Meaningful only if "value" is
    "guess[some_int]" or "muess[some_int]".

    return_offsets: If true, the function returns, after the value if also returned, a tuple of the per-dimension
    offsets of the cropped image.
    """
    if force_color:
        grayscale_mode = False
    elif isinstance(value, str):
        grayscale_mode = value[0] == GRAYSCALE_MODE_MARKER
    else:
        grayscale_mode = isinstance(value, int) or isinstance(value, float)

    if isinstance(value, str):
        method, border_width = _GuessingMethodAndBorderWidth(value)
        border_values = _BorderValues(img, method, border_width)
        if method == "guess":
            value = _ValueFromGrayscaleBorder(border_values)
        else:
            value = _ValueFromMultichannelBorder(border_values, img.shape[-1])
    elif grayscale_mode:
        if not (isinstance(value, int) or isinstance(value, float)):
            raise ValueError(f"{type(value)}: Invalid value type for grayscale")
    elif isinstance(value, int) or isinstance(value, float):
        value = img.shape[-1] * (value,)
    elif value.__len__() != img.shape[-1]:
        raise ValueError(
            f"{value.__len__()}: Invalid autocropping value length; Expected_{img.shape[-1]}"
        )

    if axis is None:
        if grayscale_mode:
            cropping_axes = tuple(range(img.ndim))
        else:
            cropping_axes = tuple(range(img.ndim - 1))
    elif isinstance(axis, int):
        cropping_axes = (axis,)
    else:
        cropping_axes = axis
    if grayscale_mode:
        kept_axes = set(range(img.ndim))
    else:
        kept_axes = set(range(img.ndim - 1))

    if tolerance == 0.0:
        vequ_bmap = _StrictValueEqualityBMap(img, value, grayscale_mode)
    elif grayscale_mode:
        vequ_bmap = _FuzzyGrayscaleValueEqualityBMap(img, value, tolerance)
    else:
        vequ_bmap = _FuzzyMultichannelValueEqualityBMap(img, value, tolerance)

    slices = img.ndim * [slice(None)]
    for cropping_axis in cropping_axes:
        kept_axes.remove(cropping_axis)
        #
        all_equal = nmpy.all(vequ_bmap, axis=tuple(kept_axes))
        padded_all_equal = nmpy.pad(all_equal, 1, constant_values=True)
        change_idc = nmpy.where(nmpy.diff(padded_all_equal) > 0)[0]
        #
        if change_idc.__len__() > 1:
            slices[cropping_axis] = slice(change_idc[0], change_idc[-1])
        else:
            slices[cropping_axis] = slice(0, 0)
        #
        kept_axes.add(cropping_axis)

    # Returned sliced image is copied to avoid modifying original image from cropped one
    output = [img[tuple(slices)].copy()]

    if return_value:
        if not isinstance(value, array_t):
            value = nmpy.array(value)
        output.append(value)

    if return_offsets:
        if grayscale_mode:
            slices_for_offsets = slices
        else:
            slices_for_offsets = slices[:-1]
        offsets = tuple(one_slice.start for one_slice in slices_for_offsets)
        output.append(offsets)

    if output.__len__() > 1:
        return tuple(output)
    else:
        return output[0]


def _GuessingMethodAndBorderWidth(value: str) -> Tuple[str, int]:
    """"""
    match = VALUE_GUESSING_REGEX.match(value)
    if match:
        groups = match.groups()
        method = groups[0]
        if groups[1].__len__() > 1:
            border_width = int(groups[1])
            if border_width < 1:
                raise ValueError(f"{border_width}: Invalid border width; Expected_>0")
        else:
            border_width = 1
    else:
        raise ValueError(f"{value}: Invalid autocropping value")

    return method, border_width


def _BorderValues(img: array_t, method: str, border_width: int) -> List[intensity_h]:
    """
    Note that some border values are repeated in the list. This is not correct, but probably not too dramatic for the
    subsequent steps.
    """
    first_border = img.ndim * [slice(None)]
    last_border = img.ndim * [slice(None)]
    first_border[0] = slice(border_width)
    last_border[0] = slice(-border_width, None)

    output = []
    if method == "guess":
        n_rolls = img.ndim
    else:
        n_rolls = img.ndim - 1
    for _ in range(n_rolls):
        output.extend(img[tuple(first_border)].flatten())
        output.extend(img[tuple(last_border)].flatten())

        first_border = [first_border[-1]] + first_border[:-1]
        last_border = [last_border[-1]] + last_border[:-1]

    return output


def _ValueFromGrayscaleBorder(border_values: List[intensity_h]) -> intensity_h:
    """
    Since some border values are repeated in the list, the result is biased. However, this has a significant influence
    only for tiny images.
    """
    unique_values = nmpy.unique(border_values)
    bin_edges = nmpy.concatenate((unique_values, (unique_values[-1] + 1,)))
    n_occurrences, _ = nmpy.histogram(border_values, bins=bin_edges)
    most_frequent = nmpy.argmax(n_occurrences)
    if n_occurrences[most_frequent] == 1:
        output = nmpy.median(border_values)
    else:
        output = unique_values[most_frequent]

    return output


def _ValueFromMultichannelBorder(
    border_values: List[intensity_h], n_channels: int
) -> spectrum_h:
    """
    Since some border values are repeated in the list, the result is biased. However, this has a significant influence
    only for tiny images.
    """
    border_values = nmpy.reshape(border_values, (n_channels, -1))
    unique_values = nmpy.unique(border_values, axis=1).T
    channel_shape = (n_channels, 1)
    n_occurrences = (
        nmpy.count_nonzero(border_values == nmpy.reshape(value, channel_shape))
        for value in unique_values
    )
    n_occurrences = nmpy.fromiter(n_occurrences, nmpy.int64)
    most_frequent = nmpy.argmax(n_occurrences)
    if n_occurrences[most_frequent] == 1:
        output = nmpy.median(border_values, axis=1)
    else:
        output = unique_values[most_frequent, :]

    return output


def _StrictValueEqualityBMap(
    img: array_t, value: pix_value_h, grayscale_mode: bool
) -> array_t:
    """"""
    if grayscale_mode:
        output = img == value
    else:
        output = img[..., 0] == value[0]
        for idx in range(1, value.__len__()):
            nmpy.logical_and(output, img[..., idx] == value[idx], out=output)

    return output


def _FuzzyGrayscaleValueEqualityBMap(
    img: array_t, value: intensity_h, tolerance: float
) -> array_t:
    """"""
    min_value, max_value, *_ = spim.extrema(img)
    full_dynamic = max_value - min_value
    half_interval = 0.5 * tolerance * full_dynamic

    lower_threshold = value - half_interval
    upper_threshold = value + half_interval
    # It is not a problem to consider the following 2 cases as mutually exclusive, although they are not. The
    # shifts are done so that if value is too close to min_value or max_value, it still gets a full tolerance
    # instead of a "truncated" one.
    if lower_threshold < min_value:
        shift = min_value - lower_threshold
        lower_threshold = min_value
        upper_threshold += shift
    elif upper_threshold > max_value:
        shift = upper_threshold - max_value
        lower_threshold -= shift
        upper_threshold = max_value

    output = nmpy.logical_and(img >= lower_threshold, img <= upper_threshold)

    return output


def _FuzzyMultichannelValueEqualityBMap(
    img: array_t, value: spectrum_h, tolerance: float
) -> array_t:
    """"""
    min_value, max_value, full_dynamic = None, None, 0.0
    for c_idx in range(img.shape[-1]):
        min_current, max_current, *_ = spim.extrema(img[..., c_idx])
        current_dynamic = max_current - min_current
        if current_dynamic > full_dynamic:
            full_dynamic = current_dynamic
    half_interval = 0.5 * tolerance * full_dynamic

    # Unfortunately, the multichannel mode does not get the full tolerance management of the grayscale mode
    value = nmpy.reshape(value, (img.ndim - 1) * (1,) + (img.ndim,))
    color_centered = img.astype(nmpy.float64) - value
    color_sq_distances = nmpy.sum(color_centered**2, axis=-1)

    output = color_sq_distances <= half_interval**2

    return output


def PaddedToCommonShape(
    imgs: Sequence[array_t],
    /,
    *,
    padding_type: str = "constant",
    type_details: Dict[str, Any] = None,
    value_combination: str = None,
    background: array_t = None,
    offsets: Sequence[Sequence[int]] = None,
) -> Tuple[array_t, ...]:
    """
    All the images must have the same dimension, and there is no distinction between single- and multiple-channel images
    """
    if type_details is None:
        type_details = {}
    elif value_combination is not None:
        # A copy is necessary (to avoid a side effect) since the details will be updated
        type_details = type_details.copy()

    all_lengths = nmpy.array(tuple(img.shape for img in imgs), dtype=nmpy.int64)
    max_lengths = nmpy.amax(all_lengths, axis=0, keepdims=True)
    total_paddings = max_lengths - all_lengths
    paddings_before = nmpy.floor_divide(total_paddings, 2)
    paddings_after = total_paddings - paddings_before

    if background is None:
        if value_combination is not None:
            if value_combination == "min":
                CombinedValue = nmpy.amin
            elif value_combination == "max":
                CombinedValue = nmpy.amax
            elif value_combination == "mean":
                CombinedValue = nmpy.mean
            elif value_combination == "median":
                CombinedValue = nmpy.median
            else:
                raise ValueError(f"{value_combination}: Invalid value combination")
            if padding_type == "min":
                PaddingValue = nmpy.amin
            elif padding_type == "max":
                PaddingValue = nmpy.amax
            elif padding_type == "mean":
                PaddingValue = nmpy.mean
            elif padding_type == "median":
                PaddingValue = nmpy.median
            else:
                raise ValueError(
                    f"{padding_type}: Invalid padding type when combining values"
                )
            padding_type = "constant"
            padding_value = CombinedValue(tuple(PaddingValue(img) for img in imgs))
            type_details["constant_values"] = padding_value
    elif nmpy.any(max_lengths != background.shape):
        raise ValueError(
            f"{background.shape}: Invalid background lengths; Expected_{tuple(nmpy.squeeze(max_lengths))}"
        )

    output = []
    dimension_idc = tuple(range(imgs[0].ndim))
    for i_idx, img in enumerate(imgs):
        # Paddings after are equal to padding before + 0 or 1
        if nmpy.any(paddings_after[i_idx, :]):
            per_dim_paddings = tuple(
                (paddings_before[i_idx, d_idx], paddings_after[i_idx, d_idx])
                for d_idx in dimension_idc
            )
            if background is None:
                padded = nmpy.pad(
                    img, per_dim_paddings, mode=padding_type, **type_details
                )
            else:
                slices = img.ndim * [slice(None)]
                img_shape = img.shape
                if offsets is None:
                    for d_idx, paddings in enumerate(per_dim_paddings):
                        slices[d_idx] = slice(
                            paddings[0], (paddings[0] + img_shape[d_idx])
                        )
                else:
                    for d_idx, offset in enumerate(offsets[i_idx]):
                        slices[d_idx] = slice(offset, (offset + img_shape[d_idx]))
                padded = background.astype(img.dtype)
                padded[tuple(slices)] = img
            output.append(padded)
        else:
            output.append(img.copy())

    return tuple(output)



# Copyright CNRS/Inria/UCA
# Contributor(s): Eric Debreuve (since 2019)
#
# eric.debreuve@cnrs.fr
#
# This software is governed by the CeCILL  license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".
#
# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.
#
# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.
#
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

import importlib as mprt
import itertools as ittl
from types import ModuleType as module_t
from typing import Sequence, Tuple

import imageio as mgio
import numpy as nmpy
from im_tools_36.path import path_h, path_t


def _Module(name: str, /) -> module_t | None:
    """"""
    try:
        output = mprt.import_module(name)
    except ModuleNotFoundError:
        output = None

    return output


aics = _Module("aicsimageio")
mrci = _Module("mrc")
skio = _Module("skimage.io")
tiff = _Module("tifffile")


array_t = nmpy.ndarray


def _WithAICS(path: str, /) -> array_t:
    """
    Arrangement: TCZYX
    """
    return aics.AICSImage(path).data


def _WithMRC(path: str, /) -> array_t:
    """
    If output.ndim == 5, probably time x channel x Z x Y x X, while sequences are:
        time x channel   x (Z=1 x)             Y x X.
    So one gets:
        time x channel=1 x Z=actual channels x Y x X. Then, use: output[:, 0, :, :]
    numpy.array: Because the returned value seems to be a read-only memory map
    """
    return nmpy.array(mrci.imread(path))


READING_MODULE = {
    "aicsimageio": aics,
    "imageio": mgio,
    "mrc": mrci,
    "scikit-image": skio,
    "skimage": skio,
    "tifffile": tiff,
}
READING_FUNCTION = {
    aics: _WithAICS,
    mgio: mgio.v3.imread,
    mrci: _WithMRC,
    skio: skio.imread,
    tiff: tiff.imread,
    None: None,
}


def ImageVolumeOrSequence(
    path: path_h,
    /,
    *,
    with_module: str = None,
    should_squeeze: bool = True,
    expected_dim: int = None,
    expected_shape: Sequence[int | None] = None,
) -> array_t | Tuple[str, ...]:
    """"""
    # Potential outputs
    image = None
    issues = []

    if isinstance(path, str):
        path_str = path
        path_lib = path_t(path)
    else:
        path_str = str(path)
        path_lib = path
    if (expected_dim is None) and (expected_shape is not None):
        expected_dim = expected_shape.__len__()

    if with_module is None:
        reading_functions = [READING_FUNCTION[mgio], READING_FUNCTION[skio]]

        img_format = path_lib.suffix[1:].lower()
        if img_format in ("tif", "tiff"):
            reading_functions.append(READING_FUNCTION[aics])
            reading_functions.append(READING_FUNCTION[tiff])
        elif img_format in ("dv", "mrc"):
            if img_format == "dv":
                reading_functions.append(READING_FUNCTION[aics])
            reading_functions.append(READING_FUNCTION[mrci])
        elif img_format in ("czi", "lif", "nd2"):
            reading_functions.append(READING_FUNCTION[aics])
    elif with_module in READING_MODULE:
        reading_functions = [READING_FUNCTION[READING_MODULE[with_module]]]
    else:
        return (
            f"{with_module}: Invalid module. Expected={str(tuple(READING_MODULE.keys()))[1:-1]}",
        )

    for Read in reversed(reading_functions):
        if Read is None:
            continue

        try:
            image = Read(path_str)
            break
        except Exception as exception:
            issues.append(str(exception))

    if image is None:
        return tuple(issues)

    if should_squeeze:
        image = nmpy.squeeze(image)

    if (expected_dim is not None) and (image.ndim != expected_dim):
        return (
            f"{image.ndim}: Invalid dimension (shape={image.shape}). Expected={expected_dim}",
        )
    if expected_shape is None:
        return image

    shape = image.shape
    if _ShapesMatch(shape, expected_shape):
        return image

    shape_as_array = nmpy.array(shape)
    for order in ittl.permutations(range(image.ndim)):
        if _ShapesMatch(shape_as_array[nmpy.array(order)], expected_shape):
            return nmpy.moveaxis(image, order, range(image.ndim))

    return (
        f"{shape}: Invalid shape. Expected={tuple(expected_shape)}, or a permutation of it",
    )


def _ShapesMatch(
    actual: Tuple[int, ...] | array_t, expected: Sequence[int | None], /
) -> bool:
    """"""
    return all((_ctl == _xpt) or (_xpt is None) for _ctl, _xpt in zip(actual, expected))



# Copyright CNRS/Inria/UCA
# Contributor(s): Eric Debreuve (since 2019)
#
# eric.debreuve@cnrs.fr
#
# This software is governed by the CeCILL  license under French law and
# abiding by the rules of distribution of free software.  You can  use,
# modify and/ or redistribute the software under the terms of the CeCILL
# license as circulated by CEA, CNRS and INRIA at the following URL
# "http://www.cecill.info".
#
# As a counterpart to the access to the source code and  rights to copy,
# modify and redistribute granted by the license, users are provided only
# with a limited warranty  and the software's author,  the holder of the
# economic rights,  and the successive licensors  have only  limited
# liability.
#
# In this respect, the user's attention is drawn to the risks associated
# with loading,  using,  modifying and/or developing or reproducing the
# software by the user in light of its specific status of free software,
# that may mean  that it is complicated to manipulate,  and  that  also
# therefore means  that it is reserved for developers  and  experienced
# professionals having in-depth computer knowledge. Users are therefore
# encouraged to load and test the software's suitability as regards their
# requirements in conditions enabling the security of their systems and/or
# data to be ensured and,  more generally, to use and operate it in the
# same conditions as regards security.
#
# The fact that you are presently reading this means that you have had
# knowledge of the CeCILL license and that you accept its terms.

from typing import Final, Sequence, Tuple, Union

import matplotlib as mplb
import matplotlib.pyplot as pypl
import mplb_tools_36.mplb_tools_36 as mp36
import numpy as nmpy
import skimage.color as sicl
import skimage.data as sidt
from im_tools_36.intensity import GloballyRelightedVersion
from im_tools_36.shape import AutoCropped, PaddedToCommonShape
from magicgui import event_loop, magicgui


# Because MagicGUI uses it, and only one backend can be sued at a time, I suppose
mplb.use("QtCairo")


GUESS_PREFIXES: Final = {2: "g", 3: "m"}


@magicgui(layout="vertical", call_button="Run")
def AutocropRocket(*, grayscale_mode: bool = False, border_width: int = 10) -> None:
    """"""
    img = sidt.rocket()
    cropping_tolerances_grayscale = nmpy.linspace(0.0, 0.95, num=8)
    cropping_tolerances_color = (0.0, 0.85, 0.95, 1.05, 1.35, 1.5, 1.8, 2.0)
    text_color = "c"

    if grayscale_mode:
        img = sicl.rgb2gray(img)
        cropping_tolerances = tuple(cropping_tolerances_grayscale)
    else:
        cropping_tolerances = tuple(cropping_tolerances_color)
    # Insert dummy tolerance for original image
    cropping_tolerances = (None,) + cropping_tolerances

    guess_mode = f"{GUESS_PREFIXES[img.ndim]}uess{border_width}"

    cropped_imgs = [img]
    cropping_values = [None]
    cropped_offsets = [None]
    for tolerance in cropping_tolerances[1:]:
        tolerance: float
        cropped, value, offsets = AutoCropped(
            img,
            value=guess_mode,
            tolerance=tolerance,
            return_value=True,
            return_offsets=True,
        )
        cropped_imgs.append(cropped)
        cropping_values.append(value.tolist())
        cropped_offsets.append(offsets)

    if grayscale_mode:
        background = 0.4 * img
    else:
        background = GloballyRelightedVersion(img, -25)
    p_cropped_imgs = PaddedToCommonShape(
        cropped_imgs, background=background, offsets=cropped_offsets
    )

    _, all_axes = mp36.SimpleDisplay(
        p_cropped_imgs, auto_grid=True, mode="make_only", return_axes=True
    )
    for a_idx, axes in enumerate(all_axes):
        if axes is None:
            continue

        cropping_value_as_str = _CroppingValueAsStr(cropping_values[a_idx])
        cropping_tolerance_as_str = _CroppingToleranceAsStr(cropping_tolerances[a_idx])
        lengths_as_str = _LengthsAsStr(cropped_imgs[a_idx].shape, img.shape[:2])

        mp36.RemoveColorbarFromAxes(axes)
        axes.set_axis_off()
        axes.annotate(
            f"{cropping_value_as_str}{cropping_tolerance_as_str}={lengths_as_str}",
            xy=(4, 8),
            color=text_color,
            verticalalignment="top",
            fontsize="x-small",
        )

    pypl.show()


def _CroppingValueAsStr(
    cropping_value: Union[Union[int, float], Sequence[Union[int, float]]], /
) -> str:
    """"""
    if cropping_value is None:
        output = "Original Image"
    else:
        if not isinstance(cropping_value, Sequence):
            cropping_value = (cropping_value,)

        elms_as_str = []
        for elm in cropping_value:
            if float(elm).is_integer():
                elms_as_str.append(int(elm).__str__())
            else:
                elms_as_str.append(f"{elm:.2f}")

        if elms_as_str.__len__() > 1:
            output = "(" + ",".join(elms_as_str) + ")"
        else:
            output = elms_as_str[0]

    return output


def _CroppingToleranceAsStr(cropping_tolerance: float | None, /) -> str:
    """"""
    if cropping_tolerance is None:
        output = ""
    else:
        output = f"+{cropping_tolerance:.2f}"

    return output


def _LengthsAsStr(lengths: Tuple[int, ...], ref_lengths: Tuple[int, ...], /) -> str:
    """"""
    height, width, *_ = lengths
    if (height, width) == ref_lengths:
        output = "Original Size"
    else:
        output = f"{width}x{height}"

    return output


if __name__ == "__main__":
    #
    with event_loop():
        _ = AutocropRocket.show()
